#pragma once

#include <stdint.h>  // IWYU pragma: keep
#include <stdlib.h>  // IWYU pragma: keep

#include "nvim/tui/termkey/termkey_defs.h"  // IWYU pragma: keep

#include "tui/termkey/termkey.h.generated.h"
