---@diagnostic disable: no-unknown
if arg[1] == '--help' then
  print('Usage:')
  print('  ' .. arg[0] .. ' [-c] target source varname [source varname]...')
  print('')
  print('Generates C file with big uint8_t blob.')
  print('Blob will be stored in a static const array named varname.')
  os.exit()
end

-- Recognized options:
--   -c   compile Lua bytecode
local options = {}

local ignorelist = {
  ['vim._core.editor'] = true,
  ['vim._core.options'] = true,
  ['vim.keymap'] = true,
}

while true do
  local opt = string.match(arg[1], '^-(%w)')
  if not opt then
    break
  end

  options[opt] = true
  table.remove(arg, 1)
end

assert(#arg >= 3 and (#arg - 1) % 2 == 0)

local target_file = arg[1] or error('Need a target file')
local target = assert(io.open(target_file, 'w'))

target:write('#include <stdint.h>\n\n')

local index_items = {}

local modnames = {}
for argi = 2, #arg, 2 do
  local source_file = arg[argi]
  local modname = arg[argi + 1]
  if modnames[modname] then
    error(string.format('modname %q is already specified for file %q', modname, modnames[modname]))
  end
  modnames[modname] = source_file

  local varname = string.gsub(modname, '%.', '_dot_') .. '_module'
  target:write(('static const uint8_t %s[] = {\n'):format(varname))

  local source = io.open(source_file, 'r')
    or error(string.format("source_file %q doesn't exist", source_file))
  local output = source:read('*a')
  source:close()
  if options.c then
    local prefix = ignorelist[modname] and '' or '@'
    local relpath = modname:gsub('%.', '/')
    output = string.dump(assert((loadstring or load)(output, prefix .. relpath)), false)
  end

  local num_bytes = 0
  local MAX_NUM_BYTES = 15 -- 78 / 5: maximum number of bytes on one line
  target:write(' ')

  local increase_num_bytes
  increase_num_bytes = function()
    num_bytes = num_bytes + 1
    if num_bytes == MAX_NUM_BYTES then
      num_bytes = 0
      target:write('\n ')
    end
  end

  for i = 1, string.len(output) do
    local byte = output:byte(i)
    target:write(string.format(' %3u,', byte))
    increase_num_bytes()
  end

  target:write('  0};\n')
  if modname ~= '_' then
    table.insert(
      index_items,
      '  { "' .. modname .. '", ' .. varname .. ', sizeof ' .. varname .. ' },\n\n'
    )
  end
end

target:write('static ModuleDef builtin_modules[] = {\n')
target:write(table.concat(index_items))
target:write('};\n')

target:close()
