#include <fcntl.h>

static const mode_t kO_RDONLY = O_RDONLY;
static const mode_t kO_WRONLY = O_WRONLY;
static const mode_t kO_RDWR   = O_RDWR;
static const mode_t kO_CREAT  = O_CREAT;
static const mode_t kO_EXCL   = O_EXCL;
